'''
Jacob Thompson

MakeSet.py
'''
import numpy as np

def Line(m, b, num = 50, noise_deviation = 1, span = [-10, 10]):
	'''
	Makes a set of x's and y's in the shape of a parabola following the equation:
		y = m * x + b
	
	Adds noise and returns both x and y
	
	Note: The returned values are sorted in ascending order by the x's
	'''
	x = np.random.rand(num, 1) * (span[1] - span[0]) + span[0]
	x[:, 0].sort()
	noise = np.random.randn(num, 1) * noise_deviation
	y = x @ [[m]] + b + noise
	return x, y
	
def Parabola(a, b, c, num = 50, noise_deviation = 1, span = [-10, 10]):
	'''
		Makes a set of x's and y's shaped in a parabola of the form:
			y = ax^2 + bx + c
			
		Adds noise and returns both x and y
	'''
	x = np.random.rand(num, 1) * (span[1] - span[0]) + span[0]	# Generates random smattering of x's
	x[:, 0].sort()												# Sort the x's
	noise = np.random.randn(num, 1) * noise_deviation			# Generate a column of random noise
	X = np.hstack((np.ones((num, 1)), x, x ** 2))				# Generate the X array by combining ones, x, and x^2
	w = [[c], [b], [a]]											# Builds the weight vector from a, b, and c
	y = X @ w + noise											# Calculates y
	
	return x, y
	
def MultiLine(weights, num = 50, noise_deviation = 1, span = [-10, 10]):
	if type(weights) != np.array:
		weights = np.array(weights)
	
	dims = weights.shape[0] - 1
	# print(dims)
	x = np.random.rand(num, dims) * (span[1] - span[0]) + span[0]
	bias = np.ones([num, 1])
	x = np.hstack((bias, x))
	noise = np.random.randn(num, 1) * noise_deviation
	# print('x:', x.shape)
	# print('w:', weights.shape)
	y = x @ weights + noise	
	# print('y:', y.shape)
	return x, y



'''
Jacob Thompson

LossFunctions.py
'''
import numpy as np
import JT.ActivationFunctions

class Loss():
	'''
	Parent class that holds aliases
	'''
	def __init_subclass__(cls):
		cls.Eval = cls.Evaluate
		cls.__call__ = cls.Evaluate
		
		cls.D = cls.Derivative

class SumOfSquaredResidual():
	'''
	Class that contains the formula for calculating the Sum of Squared Residual (aka Sum of Squared Error)
	'''
	def __init__(self):
		'''
		Initializes the object
		
		Note: Nothing really goes on here, but this will be the format for other loss functions
		'''
		self.name = 'Sum of Squared Residuals'
		
	def Evaluate(self, w_hat, x, y, y_hat):
		'''
		Evaluates the loss function given all the arguments
		
		Note: 	w_hat and x are unused, but they are used in other loss functions so we keep them as 
				input parameters for consistent formatting
		'''
		return ((y - y_hat).T @ (y - y_hat)).trace()
	
	def Derivative(self, w_hat, x, y, y_hat):
		'''
		Evaluates the derivative of the loss function given all the arguments
		
		Note: 	w_hat and x are unused, but they are used in other loss functions so we keep them as 
				input parameters for consistent formatting
		'''
		return x.T @ (y_hat - y)
		
class SumOfSquaredError(SumOfSquaredResidual):
	'''
	Alias for SumOfSquaredResidual
	'''
	pass
		
class L1SumOfSquaredResidual():
	'''
	Class that contains the formula for calculating the L1 Penalized Sum of Squared Residual (aka Sum of Squared Error)
	'''
	def __init__(self, lamb):
		'''
		Initializes the object and stores the parameter for lambda - the penalty modifier
		'''
		self.name = 'L1 Penalized Sum of Squared Residuals'
		self.lamb = lamb
		
	def Evaluate(self, w_hat, x, y, y_hat):
		'''
		Evaluates the loss function given all the arguments
		
		Note: 	x is unused, but it is used in other loss functions so we keep it as 
				an input parameter for consistent formatting
		'''
		return (
			((y - y_hat).T @ (y - y_hat)).trace() + 	# This is the formula for the typical Sum of Squared Residual
			self.lamb * np.sum(abs(w_hat))				# This is the formula for the L1 penalty
			)
	
	def Derivative(self, w_hat, x, y, y_hat):
		'''
		Evaluates the derivative of the loss function given all the arguments
		'''
		return (
			x.T @ (y_hat - y) + 			# This is the formula for the typical Sum of Squared Residual derivative
			self.lamb * np.sign(w_hat)		# This is the formula for the L1 penalty derivative
			)
			
class L1SumOfSquaredError(L1SumOfSquaredResidual):
	'''
	Alias for L1SumOfSquaredResidual
	'''
	pass
		
class L2SumOfSquaredResidual():
	'''
	Class that contains the formula for calculating the L2 Penalized Sum of Squared Residual (aka Sum of Squared Error)
	'''
	def __init__(self, lamb):
		'''
		Initializes the object and stores the parameter for lambda - the penalty modifier
		'''
		self.name = 'L2 Penalized Sum of Squared Residuals'
		self.lamb = lamb
		
	def Evaluate(self, w_hat, x, y, y_hat):
		'''
		Evaluates the loss function given all the arguments
		
		Note: 	x is unused, but it is used in other loss functions so we keep it as 
				an input parameter for consistent formatting
		'''
		return (
			((y - y_hat).T @ (y - y_hat)).trace() + 	# This is the formula for the typical Sum of Squared Residual
			self.lamb * float(w_hat.T @ w_hat)			# This is the formula for the L2 penalty
			)
	
	def Derivative(self, w_hat, x, y, y_hat):
		'''
		Evaluates the derivative of the loss function given all the arguments
		'''
		return (
			x.T @ (y_hat - y) + 	# This is the formula for the typical Sum of Squared Residual derivative
			self.lamb * w_hat		# This is the formula for the L2 penalty derivative
			)

class L2SumOfSquaredError(L2SumOfSquaredResidual):
	'''
	Alias for L2SumOfSquaredResidual
	'''
	pass

			
class L1L2SumOfSquaredResidual():
	'''
	Class that contains the formula for calculating the L1 and L2 Penalized Sum of Squared Residual (aka Sum of Squared Error) (aaka ElasticNet)
	'''
	def __init__(self, lamb_1 = None, lamb_2 = None, reg_rate = None, l1_ratio = None):
		'''
		Initializes the object and stores the parameter for:
			lambda_1 - the penalty modifier for L1 penalization
			lambda_2 - the penalty modifier for L2 penalization
		'''
		if reg_rate and l1_ratio:
			self.lamb_1 = reg_rate * l1_ratio
			self.lamb_2 = reg_rate * (1-l1_ratio)
		elif lamb_1 and lamb_2:
			self.lamb_1 = lamb_1
			self.lamb_2 = lamb_2
		self.name = 'L1 and L2 Penalized Sum of Squared Residuals'
		
		
	def Evaluate(self, w_hat, x, y, y_hat):
		'''
		Evaluates the loss function given all the arguments
		
		Note: 	x is unused, but it is used in other loss functions so we keep it as 
				an input parameter for consistent formatting
		'''
		return (
			((y - y_hat).T @ (y - y_hat)).trace() + 	# This is the formula for the typical Sum of Squared Residual
			self.lamb_1 * np.sum(abs(w_hat)) + 			# This is the formula for the L1 penalty
			self.lamb_2 * (w_hat.T @ w_hat).trace()		# This is the formula for the L2 penalty
			)
	
	def Derivative(self, w_hat, x, y, y_hat):
		'''
		Evaluates the derivative of the loss function given all the arguments
		'''
		return (
			x.T @ (y_hat - y) + 				# This is the formula for the typical Sum of Squared Residual derivative
			self.lamb_1 * np.sign(w_hat) + 		# This is the formula for the L1 penalty derivative
			self.lamb_2 * w_hat					# This is the formula for the L2 penalty derivative
			)
			
class L1L2SumOfSquaredError(L1L2SumOfSquaredResidual):
	'''
	Alias for L1L2SumOfSquaredResidual
	'''
	pass
			
class ElasticNet(L1L2SumOfSquaredResidual):
	'''
	Alias for L1L2SumOfSquaredResidual
	'''
	pass
	
class Binary_Cross_Entropy():
	def __init__(self, reg_rate = 0, l1_ratio = 0):
		self.name = 'Binary Cross Entropy'
		self.lamb_1 = reg_rate * l1_ratio
		self.lamb_2 = reg_rate * (1-l1_ratio)
		self.activation = JT.ActivationFunctions.Sigmoid()
	
	def Evaluate(self, w_hat, x, y, y_hat):
		p_hat = self.activation(y_hat)
		p_hat += (p_hat < 0.001) * 0.001
		p_hat -= (p_hat > 0.999) * 0.001
			
		return (
			-np.sum(y * np.log(p_hat) + (1 - y) * np.log(1 - p_hat)) 
			# self.lamb_1 * np.sum(abs(w_hat)) + 								# This is the formula for the L1 penalty
			# self.lamb_2 * (w_hat.T @ w_hat).trace()	
			)
		
	def Derivative(self, w_hat, x, y, y_hat):
		p_hat = self.activation(y_hat)
		p_hat += (p_hat < 0.001) * 0.001
		p_hat -= (p_hat > 0.999) * 0.001
		
		return (
			x.T @ (p_hat - y) + 
			self.lamb_1 * np.sign(w_hat) + 						# This is the formula for the L1 penalty derivative
			self.lamb_2 * w_hat	
			)
	
class Cross_Entropy():
	'''
	==== WIP ====
	'''
	def __init__(self, reg_rate = 0, l1_ratio = 0, activation = JT.ActivationFunctions.Sigmoid()):
		self.lamb_1 = reg_rate * l1_ratio
		self.lamb_2 = reg_rate * (1-l1_ratio)
		self.activation = activation
		
	def Evaluate(self, w_hat, x, y, y_hat):
		p_hat = self.activation(y_hat)
		print(p_hat)
		return (
			-np.sum(y * np.log(p_hat) + (1 - y) * np.log(1 - p_hat)) + 		
			self.lamb_1 * np.sum(abs(w_hat)) + 								# This is the formula for the L1 penalty
			self.lamb_2 * (w_hat.T @ w_hat).trace()							# This is the formula for the L2 penalty
			)
			
	def Derivative(self, w_hat, x, y, y_hat):
		p_hat = self.activation(y_hat)
		print(p_hat)
		return (
			-x.T @ (y * self.activation.D(y_hat) / p_hat) + 
			self.lamb_1 * np.sign(w_hat) + 						# This is the formula for the L1 penalty derivative
			self.lamb_2 * w_hat									# This is the formula for the L2 penalty derivative
			)


'''
Jacob Thompson

Activation Functions
'''
import numpy as np

class Activation():		
	def __init_subclass__(cls):
		cls.Eval = cls.Evaluate
		cls.__call__ = cls.Evaluate
		
		cls.D = cls.Derivative
	
class Sigmoid(Activation):
	def Evaluate(self, z):
		'''
		Evaluates the sigmoid of z:
			1 / (1 + e^(-z))
		'''
		return 1 / (1 + np.exp(-z))
		
	def Derivative(self, z):
		'''
		Evaluates the derivative of sigmoid of z:
			Sigmoid(z) * (1 - Sigmoid(z))
		'''
		sig = self.Evaluate(z)
		return sig * (1 - sig)
		

